@@ -1,7 +1,3 @@
-#ifndef SQUID_CONFIG_H
-#include "config.h"
-#endif
-
 #ifndef SQUID_OS_SUNOS_H
 #define SQUID_OS_SUNOS_H
 
@@ -14,6 +10,11 @@
  ****************************************************************************/
 
 
+/* O_NONBLOCK requires <fcntl.h> to be included first */
+#if HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+
 /*
  * We assume O_NONBLOCK is broken, or does not exist, on SunOS.
  */
