@@ -37,11 +37,12 @@
 
 #include "config.h"
 
-#if USE_DELAY_POOLS
+#if USE_DELAY_POOLS && USE_AUTH
 #include "squid.h"
 #include "DelayUser.h"
 #include "auth/UserRequest.h"
 #include "auth/User.h"
+#include "comm/Connection.h"
 #include "NullDelayId.h"
 #include "Store.h"
 
@@ -235,4 +236,4 @@ DelayUser::Id::bytesIn(int qty)
     theBucket->theBucket.bytesIn(qty);
 }
 
-#endif
+#endif /* USE_DELAY_POOLS && USE_AUTH */
