@@ -34,6 +34,8 @@
 
 #include "squid.h"
 #include "comm.h"
+#include "comm/Connection.h"
+#include "comm/Loops.h"
 #include "CommRead.h"
 #include "fde.h"
 
@@ -60,20 +62,20 @@ DeferredReadManager::kickReads(int const count)
 }
 
 void
-comm_read(int fd, char *buf, int size, IOCB *handler, void *handler_data)
+comm_read(const Comm::ConnectionPointer &conn, char *buf, int size, IOCB *handler, void *handler_data)
 {
     fatal ("Not implemented");
 }
 
 void
-comm_read(int, char*, int, AsyncCall::Pointer &callback)
+comm_read(const Comm::ConnectionPointer &conn, char*, int, AsyncCall::Pointer &callback)
 {
     fatal ("Not implemented");
 }
 
 /* should be in stub_CommRead */
 #include "CommRead.h"
-CommRead::CommRead (int fd, char *buf, int len, AsyncCall::Pointer &callback)
+CommRead::CommRead(const Comm::ConnectionPointer &, char *buf, int len, AsyncCall::Pointer &callback)
 {
     fatal ("Not implemented");
 }
@@ -89,18 +91,19 @@ commSetCloseOnExec(int fd)
     /* for tests... ignore */
 }
 
+#if 0
 void
-commSetSelect(int fd, unsigned int type, PF * handler, void *client_data,
-              time_t timeout)
+Comm::SetSelect(int fd, unsigned int type, PF * handler, void *client_data, time_t timeout)
 {
     /* all test code runs synchronously at the moment */
 }
 
 void
-comm_quick_poll_required()
+Comm::QuickPollRequired()
 {
     /* for tests ... ignore */
 }
+#endif
 
 int
 ignoreErrno(int ierrno)
@@ -130,15 +133,6 @@ commUnsetNonBlocking(int fd)
     return -1;
 }
 
-/* bah, cheating on stub count */
-
-pid_t
-ipcCreate(int type, const char *prog, const char *const args[], const char *name, Ip::Address &local_addr, int *rfd, int *wfd, void **hIpc)
-{
-    fatal ("Not implemented");
-    return -1;
-}
-
 void
 comm_init(void)
 {
@@ -156,3 +150,23 @@ _comm_close(int fd, char const *file, int line)
 {
     fatal ("Not implemented");
 }
+
+int
+commSetTimeout(int fd, int timeout, AsyncCall::Pointer& callback)
+{
+    fatal ("Not implemented");
+    return -1;
+}
+
+int
+comm_open_uds(int sock_type, int proto, struct sockaddr_un* addr, int flags)
+{
+    fatal ("Not implemented");
+    return -1;
+}
+
+void
+comm_write(int fd, const char *buf, int size, AsyncCall::Pointer &callback, FREE * free_func)
+{
+    fatal ("Not implemented");
+}
